﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class BinaryOperator : MVRScript
    {
		protected const string OP_AND = "AND";
		protected const string OP_OR  = "OR";
		protected const string OP_XOR = "XOR";
		protected const string OP_GT  = "Greater Than";
		protected const string OP_ET  = "Equal To";
		protected const string OP_LT  = "Less Than";
		protected static List<string> m_Operations = new List<string>{OP_AND, OP_OR, OP_XOR, OP_GT, OP_ET, OP_LT};

		protected const int NUM_INPUTS = 8;

		protected JSONStorableBool[]        m_Inputs;
		protected JSONStorableStringChooser m_Operation;
		protected JSONStorableFloat         m_Count;
		protected JSONStorableBool          m_Output;
		protected OnOffTrigger              m_OutputTrigger;


        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ValueDelay</b></size>\nLogicBrick that performs binary operation on a range of inputs, e.g., AND, OR, XOR.\n\n</color>",
				220.0f, true
			);

			m_Inputs = new JSONStorableBool[NUM_INPUTS];
			for (int i=0; i<NUM_INPUTS; i++)
			{
				m_Inputs[i] = Utils.SetupToggle(this, $"Input {i+1}", false, false);
				m_Inputs[i].setCallbackFunction += UpdateValues;
			}
			
			m_Operation = Utils.SetupStringChooser(this, "Operation", m_Operations, true);
			m_Operation.setCallbackFunction += UpdateOperation;
			RegisterStringChooser(m_Operation);
			CreatePopup(m_Operation, false);

			m_Count = Utils.SetupSliderInt(this, "Count", NUM_INPUTS / 2, 0, NUM_INPUTS, true);
			m_Count.setCallbackFunction += UpdateCount;

			m_Output = new JSONStorableBool("Output", false);
			UIDynamicToggle toggle = CreateToggle(m_Output, true);
			toggle.toggle.interactable = false;
			
			m_OutputTrigger = new OnOffTrigger(this, "OnOutput");
			Utils.SetupButton(this, "Actions OnOutput", m_OutputTrigger.OpenPanel, true);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			m_OutputTrigger.SyncAtomNames();		
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[m_OutputTrigger.Name] = m_OutputTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				m_OutputTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			m_OutputTrigger.Remove();
		}

		protected void UpdateValues(bool _)
		{
			bool output = m_Output.val;
			int  count  = 0;
			for (int i=0; i<NUM_INPUTS; i++) { if (m_Inputs[i].val) { count++; } }

			switch (m_Operation.val)
			{
				case OP_AND: output = true;  for (int i=0; i<NUM_INPUTS; i++) { output &= m_Inputs[i].val; } break;
				case OP_OR : output = false; for (int i=0; i<NUM_INPUTS; i++) { output |= m_Inputs[i].val; } break;
				case OP_XOR: output = false; for (int i=0; i<NUM_INPUTS; i++) { output ^= m_Inputs[i].val; } break;
				case OP_GT : output = count >  m_Count.val; break;
				case OP_ET : output = count == m_Count.val; break;
				case OP_LT : output = count <  m_Count.val; break;
				default    : break;
			}
			m_Output.val = output;
			m_OutputTrigger.active = output;
			m_OutputTrigger.Update();
		}

		protected void UpdateOperation(string _)
		{
			UpdateValues(false);
		}

		protected void UpdateCount(float _)
		{
			UpdateValues(false);
		}
    }
}

