﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class FloatOperator : MVRScript
    {
		protected const string OP_MIN  = "Minimum";
		protected const string OP_MAX  = "Maximum";
		protected const string OP_AVG  = "Average";
		protected const string OP_DIFF = "Difference";
		protected static List<string> m_Operations = new List<string>{OP_MIN, OP_MAX, OP_AVG, OP_DIFF};

		protected JSONStorableFloat         m_ValueIn1;
		protected JSONStorableFloat         m_ValueIn2;
		protected JSONStorableStringChooser m_Operation;
		protected JSONStorableFloat         m_ValueOut;
		protected FloatTrigger              m_ValueOutTrigger;


        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>FloatOperator</b></size>\nLogicBrick that performs operations on float values from two sources, e.g., min, max, average, difference.\n\n</color>",
				220.0f, true
			);

			m_ValueIn1 = Utils.SetupSliderFloatWithRange(this, "Input Value 1", 0f, 0f, 1f, false);
			m_ValueIn1.setCallbackFunction += UpdateValues;
			m_ValueIn2 = Utils.SetupSliderFloatWithRange(this, "Input Value 2", 0f, 0f, 1f, false);
			m_ValueIn2.setCallbackFunction += UpdateValues;

			m_Operation = new JSONStorableStringChooser("Operation", m_Operations, OP_MIN, "Operation");
			m_Operation.setCallbackFunction += UpdateOperation;
			RegisterStringChooser(m_Operation);
			CreatePopup(m_Operation, false);

			m_ValueOut = new JSONStorableFloat("Output Value", 0f, 0f, 1f);
			m_ValueOut.constrained = false;
			UIDynamicSlider slider = CreateSlider(m_ValueOut, false);
			slider.quickButtonsEnabled  = false;
			slider.defaultButtonEnabled = false;
			slider.rangeAdjustEnabled   = false;
			slider.sliderControl.slider.interactable = false;
			
			m_ValueOutTrigger = new FloatTrigger(this, "OnValueOut");
			Utils.SetupButton(this, "Actions OnValueOut", m_ValueOutTrigger.OpenPanel, false);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			m_ValueOutTrigger.SyncAtomNames();		
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[m_ValueOutTrigger.Name] = m_ValueOutTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				m_ValueOutTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			m_ValueOutTrigger.Remove();
		}

		protected void UpdateValues(float _)
		{
			switch (m_Operation.val)
			{
				case OP_MIN:  m_ValueOut.val = Mathf.Min(    m_ValueIn1.val, m_ValueIn2.val); break;
				case OP_MAX:  m_ValueOut.val = Mathf.Max(    m_ValueIn1.val, m_ValueIn2.val); break;
				case OP_AVG:  m_ValueOut.val = (m_ValueIn1.val + m_ValueIn2.val) / 2f; break;
				case OP_DIFF: m_ValueOut.val = Mathf.Abs(m_ValueIn1.val - m_ValueIn2.val); break;
				default: break;
			}
			m_ValueOutTrigger.Trigger(m_ValueOut.val);
			m_ValueOutTrigger.Update();
		}

		protected void UpdateOperation(string _)
		{
			UpdateValues(0);
		}
    }
}
