﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class HysteresisThreshold : MVRScript
    {
		private JSONStorableBool  myAcceptTrigger;
		private JSONStorableFloat myThresholdValue;
		private JSONStorableFloat myHysteresisValue;
		private JSONStorableFloat myValue;
		private OnOffTrigger      myEventTrigger;
		private float myPreviousValue;
		private JSONStorableFloat myEffectiveThreshold;
		
		
        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>HysteresisThreshold</b></size>\nLogicBrick that triggers when a value crosses a specific threshold with a hysteresis applied.</color>",
				200.0f, true
			);
			
			myValue = Utils.SetupSliderFloatWithRange(this, "Value", 0, 0, 1, false);
			myValue.constrained = false;
			myValue.setCallbackFunction += Trigger;
			myPreviousValue = myValue.val;
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);
			
			myThresholdValue = Utils.SetupSliderFloatWithRange(this, "Threshold Value", 0.5f, 0, 1, false);
			myThresholdValue.constrained = false;
						
			myHysteresisValue = Utils.SetupSliderFloatWithRange(this, "Hysteresis Value", 0.1f, 0, 1, false);
			myHysteresisValue.constrained = false;
			
			myEffectiveThreshold = new JSONStorableFloat("Effective Threshold", 0.0f, 0.0f, 1.0f);
			UIDynamicSlider slider = CreateSlider(myEffectiveThreshold, false);
			slider.quickButtonsEnabled  = false;
			slider.defaultButtonEnabled = false;
			slider.rangeAdjustEnabled   = false;
			slider.sliderControl.slider.interactable = false;
			
			myEventTrigger = new OnOffTrigger(this, "OnTrigger");
			Utils.SetupButton(this, "Actions OnTrigger", myEventTrigger.OpenPanel, false);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
			
			if (myPreviousValue < myThresholdValue.val)
			{
				myEffectiveThreshold.val = myThresholdValue.val + myHysteresisValue.val;
			}
			else
			{
				myEffectiveThreshold.val = myThresholdValue.val - myHysteresisValue.val;
			}
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			myEventTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myEventTrigger.Name] = myEventTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myEventTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myEventTrigger.Remove();
		}
		
		private void Update()
		{
			myEventTrigger.Update();
		}

		private void Trigger(float v)
		{			
			if (myAcceptTrigger.val)
			{
				if ((myPreviousValue <= myEffectiveThreshold.val) && (v > myEffectiveThreshold.val))
				{
					// just went above threshold
					myEventTrigger.On();
					// lower threshold
					myEffectiveThreshold.val = myThresholdValue.val - myHysteresisValue.val;
				}
				else if ((myPreviousValue >= myEffectiveThreshold.val) && (v < myEffectiveThreshold.val))
				{
					// just went below threshold
					myEventTrigger.Off();
					// raise threshold
					myEffectiveThreshold.val = myThresholdValue.val + myHysteresisValue.val;
				}
			}
			myPreviousValue = v;
		}
    }
}
