﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class OnLoadingActions : MVRScript
    {
		public override void Init()
		{
			CreateUI();

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRenamed;
			SuperController.singleton.onAtomAddedHandlers     += OnAtomAdded;
			SuperController.singleton.onAtomRemovedHandlers   += OnAtomRemoved;

			// register slider monitors for progress
			m_sliderMonitors = new List<SliderMonitor>();
			m_sliderMonitors.Add(new SliderMonitor(this, SuperController.singleton.loadingProgressSlider));
			m_sliderMonitors.Add(new SliderMonitor(this, SuperController.singleton.loadingProgressSliderAlt));
			m_sliderMonitors.Add(new SliderMonitor(this, ImageLoaderThreaded.singleton.progressSlider));
				
			// search for AssetLoaders
			m_cuaLoaders = new List<CustomUnityAssetLoader>();
			RescanCUA_Loaders();

			m_timerLastCheck = 0;
			m_isLoading      = false;
			m_forceOnLoading = false;
		}


		protected void CreateUI()
		{
			SimpleTriggerHandler.LoadAssets();

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>OnLoadingActions</b></size>\nLogicBrick that sends events when the loading dialog appears or disappears, e.g., to stop animations until loading has finished.\n\n</color>",
				455.0f, true
			);

			m_triggerOnLoading = new OnOffTrigger(this, "OnLoading");
			Utils.SetupButton(this, "Loading Actions", m_triggerOnLoading.OpenPanel, false);

			m_triggerOnProgress = new FloatTrigger(this, "OnLoadingProgress");
			Utils.SetupButton(this, "Progress Actions", m_triggerOnProgress.OpenPanel, false);

			m_jsonCheckInterval = Utils.SetupSliderFloat(this, "Check Interval", 0.1f, 0, 1, false);

			m_jsonMonitorCUA_Loaders = Utils.SetupToggle(this, "Monitor CUA Loaders", false, false);
			m_jsonMonitorCUA_Loaders.setCallbackFunction = (bool _) => RescanCUA_Loaders();

			JSONStorableAction actionCheckNow = new JSONStorableAction("Check Now", () => CheckNow());
			RegisterAction(actionCheckNow);

			JSONStorableAction rescanCuaLoaders = new JSONStorableAction("Rescan CUA Loaders", () => RescanCUA_Loaders());
			RegisterAction(rescanCuaLoaders);

			JSONStorableAction actionForceOnLoading = new JSONStorableAction("Force OnLoading", () => ForceOnLoading());
			RegisterAction(actionForceOnLoading);
		}


		public void CheckNow()
		{
			m_timerLastCheck = 0;
			Update();
		}
		

		public void RescanCUA_Loaders()
		{
			m_cuaLoaders.Clear();
			if (m_jsonMonitorCUA_Loaders.val)
			{
				// monitoring CUA Loaders > lets check all atoms then...
				List<Atom> atoms = SuperController.singleton.GetAtoms();
				foreach (Atom a in atoms)
				{
					CustomUnityAssetLoader[] arrCual = a.GetComponentsInChildren<CustomUnityAssetLoader>(true);
					m_cuaLoaders.AddRange(arrCual);
				}
			}
		}
		

		public void ForceOnLoading()
		{
			m_forceOnLoading = true;
			CheckNow();
		}
		

		protected void OnAtomAdded(Atom atom)
		{
			if (m_jsonMonitorCUA_Loaders.val)
			{
				CustomUnityAssetLoader[] loaders = atom.GetComponentsInChildren<CustomUnityAssetLoader>(true);
				m_cuaLoaders.AddRange(loaders);
			}
		}


		protected void OnAtomRemoved(Atom atom)
		{
			if (m_jsonMonitorCUA_Loaders.val)
			{
				List<CustomUnityAssetLoader> loaders = new List<CustomUnityAssetLoader>(atom.GetComponentsInChildren<CustomUnityAssetLoader>(true));
				m_cuaLoaders.RemoveAll(e => loaders.Contains(e));
			}
		}


		protected void OnAtomRenamed(string oldid, string newid)
		{
			m_triggerOnLoading.SyncAtomNames();
			m_triggerOnProgress.SyncAtomNames();
		}


		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[m_triggerOnLoading.Name]  = m_triggerOnLoading.GetJSON(base.subScenePrefix);
				jc[m_triggerOnProgress.Name] = m_triggerOnProgress.GetJSON(base.subScenePrefix);
			}
			return jc;
		}
		

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				m_triggerOnLoading.RestoreFromJSON( jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				m_triggerOnProgress.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
			RescanCUA_Loaders(); // in case the flag has changed
		}
		

		public void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRenamed;
			SuperController.singleton.onAtomAddedHandlers     -= OnAtomAdded;
			SuperController.singleton.onAtomRemovedHandlers   -= OnAtomRemoved;
			m_triggerOnLoading.Remove();
			m_triggerOnProgress.Remove();
		}
		

		public void Update()
		{
			// increment (don't count down) timer, so a change in the update interval has an immediate effect
			float timeSinceLastCheck = Time.unscaledTime - m_timerLastCheck;
			if (timeSinceLastCheck >= m_jsonCheckInterval.val)
			{
				// everything hinges on the loading icon and dialog (shameless copy from MacGruber's LogicBricks)
				SuperController controller  = SuperController.singleton;
				GameObject      loadingIcon = controller.loadingIcon.gameObject;
				GameObject      loadingUI   = controller.loadingUI.gameObject;
				
				bool loading = loadingIcon.activeSelf || loadingUI.activeSelf || m_forceOnLoading;
				
				if (m_jsonMonitorCUA_Loaders.val)
				{
					// check if any CUA loader has an asset selected, but it's not yet loaded
					foreach (CustomUnityAssetLoader cuaLoader in m_cuaLoaders)
					{
						if (!cuaLoader.isAssetLoaded) // check this first (faster, and more likely to be true)
						{
							// no asset loaded (yet) > is one actually selected
							JSONStorableStringChooser chs = cuaLoader.GetStringChooserJSONParam("assetName");
							if (chs != null && chs.val != "None")
							{
								loading = true;
							}
						}
					}
				}

				if (loading || (m_isLoading != loading))
				{
					float progress = 1.0f;
					foreach (SliderMonitor monitor in m_sliderMonitors)
					{
						progress = Mathf.Min(progress, monitor.GetProgress(loading));
					}
				
					// SuperController.LogMessage($"Loading status: {loading} {progress*100f}");
					m_triggerOnProgress.Trigger(progress);
					m_triggerOnLoading.active = loading;
					m_isLoading               = loading;
					m_forceOnLoading          = false;
				}
				
				m_timerLastCheck = Time.unscaledTime;
			}
		}

		
		protected class SliderMonitor
		{
			public SliderMonitor(OnLoadingActions _owner, Slider _slider)
			{
				m_owner       = _owner;
				m_slider      = _slider;
				m_oldValue    = m_slider.value;
				m_oldMaxValue = m_slider.maxValue;
			}
			
			public float GetProgress(bool _loading)
			{
				if (_loading)
				{
					if (m_progress >= 1.0f)
					{
						// just changed to loading > start from 0
						m_progress    = 0;
						m_oldValue    = m_slider.value;
						m_oldMaxValue = m_slider.maxValue;
					}
					if ((m_oldValue != m_slider.value) || (m_oldMaxValue != m_slider.maxValue) || (m_slider.value == m_slider.maxValue))
					{
						// only recalculate if slider values really change, or finished
						m_progress = m_slider.value / m_slider.maxValue;
					}
				}
				else
				{
					// guess we're finished
					m_progress = 1.0f;
				}
				return m_progress;
			}
			
			protected OnLoadingActions m_owner;
			protected Slider           m_slider;
			protected float            m_oldValue, m_oldMaxValue, m_progress;
		}


		protected JSONStorableFloat  m_jsonCheckInterval;
		protected JSONStorableBool   m_jsonMonitorCUA_Loaders;
		protected OnOffTrigger       m_triggerOnLoading;
		protected FloatTrigger       m_triggerOnProgress;

		protected bool   m_isLoading, m_forceOnLoading;
		protected float  m_timerLastCheck;

		protected List<SliderMonitor>          m_sliderMonitors;
		protected List<CustomUnityAssetLoader> m_cuaLoaders;
    }
}
