﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class ProbabilisticChoice : MVRScript
    {
		private JSONStorableStringChooser myChoiceChooser;
		private JSONStorableBool          myAvoidRepetition;
		private JSONStorableBool          myAcceptTrigger;
		private JSONStorableString        myChoiceName;

		private List<object> myChoiceUI = new List<object>();

		private class Choice
		{
			public Choice(MVRScript script, string name)
			{
				parent      = script;
				myName      = name;
				Trigger     = new EventTrigger(script, $"On{name}");
				Probability = new JSONStorableFloat($"Probability {name}", 0.5f, 0f, 10f);
				Probability.isStorable = false;
				Register();
			}
			
			public Choice(MVRScript script, string name, Choice source)
			{
				parent          = script;
				Trigger         = new EventTrigger(source.Trigger);
				Probability     = new JSONStorableFloat($"Probability {name}", 0.5f, 0f, 10f);
				Probability.isStorable = false;
				Name            = name;
				Probability.val = source.Probability.val;
				Register();
			}

			public static int SortByNameAscending(Choice a, Choice b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name
			{
				get { return myName; }
				set {
					myName = value;
					Trigger.Name = $"On{myName}";
					parent.DeregisterFloat(Probability);
					Probability.name = $"Probability {myName}";
					parent.RegisterFloat(Probability);
				}
			}

			public void Register()
			{
				// nothing to do with Trigger
				parent.RegisterFloat(Probability);
			}

			public void Deregister()
			{
				Trigger.Remove();
				parent.DeregisterFloat(Probability);
			}

			private MVRScript         parent;
			private string            myName;
			public  JSONStorableFloat Probability;
			public  EventTrigger      Trigger;
		}

		private List<Choice> myChoices = new List<Choice>();
		private int myPreviousIndex1 = -1;
		private int myPreviousIndex2 = -1;
		private bool myNeedUIRefresh = false;

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ProbabilisticChoice</b></size>\nLogicBrick that, after <i>Trigger</i> is called, selects a random <i>Choice</i> to trigger based on the probability of each choice.\n\n" +
				"Use <i>Avoid Repetition</i> to take the last one or two choices out of the selection, depending on the number of total available choices.</color>",
				300.0f, true
			);
			
			myChoiceChooser = Utils.SetupStringChooser(this, "Choice", new List<string>(), false);
			myChoiceChooser.isStorable = myChoiceChooser.isRestorable = false;
			myChoiceChooser.setCallbackFunction += UISetChoice;

			Utils.SetupTwinButton(this, "Add Choice", UIAddChoice, "Duplicate Choice", UIDuplicateChoice, false);
			Utils.SetupButton(this, "Remove Choice", UIRemoveChoice, false);
			myAvoidRepetition = Utils.SetupToggle(this, "Avoid Repetition", true, false);
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);

			Utils.SetupButton(this, "Test Trigger", TriggerAction, false);
			Utils.SetupAction(this, "Trigger", TriggerAction);

			Utils.SetupSpacer(this, 15, false);

			myChoiceName = new JSONStorableString("Name", "", UIRenameChoice);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UISetChoice(string v)
		{
			Utils.RemoveUIElements(this, myChoiceUI);

			int idx = myChoiceChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			Choice choice = myChoices[idx];

			UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Choice</b></size>", false);
			myChoiceUI.Add(optionHeader);

			myChoiceName.valNoCallback = choice.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myChoiceName, false);
			myChoiceUI.Add(nameInput);

			CreateSlider(choice.Probability, false);
			myChoiceUI.Add(choice.Probability);

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + choice.Trigger.Name, choice.Trigger.OpenPanel, false);
			myChoiceUI.Add(actionsButton);
		}
	
		private void UIAddChoice()
		{
			UIAddChoice(null);
		}

		private void UIAddChoice(Choice source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Choice" + i;
				if (!myChoiceChooser.choices.Contains(name))
				{
					Choice choice;
					if (source == null)
						choice = new Choice(this, name);
					else
						choice = new Choice(this, name, source);
					myChoices.Add(choice);
					myChoices.Sort(Choice.SortByNameAscending);

					List<string> options = myChoiceChooser.choices;
					options.Clear();
					for (int e=0; e<myChoices.Count; ++e)
						options.Add(myChoices[e].Name);
					myChoiceChooser.choices = null; // force UI sync
					myChoiceChooser.choices = options;
					myChoiceChooser.val = name;
					return;
				}
			}
			SuperController.LogError("RandomChoice: Too many options!");
		}
		
		private void UIDuplicateChoice()
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			Choice choice = idx >= 0 ? myChoices[idx] : null;
			UIAddChoice(choice);
		}

		private void UIRemoveChoice()
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			List<string> options = myChoiceChooser.choices;
			options.RemoveAt(idx);
			myChoices[idx].Deregister();
			myChoices.RemoveAt(idx);

			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;

			if (idx < options.Count)
				myChoiceChooser.val = options[idx];
			else if (options.Count > 0)
				myChoiceChooser.val = options[options.Count-1];
			else
				myChoiceChooser.val = "";
		}

		private void UIRenameChoice(string name)
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			myChoices[idx].Name = name;
			
			myChoices.Sort(Choice.SortByNameAscending);
			List<string> options = myChoiceChooser.choices;
			options.Clear();
			for (int e=0; e<myChoices.Count; ++e)
				options.Add(myChoices[e].Name);
			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;
			myChoiceChooser.val = name;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray clist = new JSONArray();
				for (int i=0; i<myChoices.Count; ++i)
				{
					Choice choice = myChoices[i];
					JSONClass cc = new JSONClass();
					cc["Name"] = choice.Name;
					cc["Probability"].AsFloat = choice.Probability.val;
					cc[choice.Trigger.Name] = choice.Trigger.GetJSON(base.subScenePrefix);
					clist.Add("", cc);
				}
				jc["Choices"] = clist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myChoices.Count; ++i)
					myChoices[i].Deregister();
				myChoices.Clear();

				JSONArray clist = jc["Choices"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass cc = clist[i].AsObject;
					Choice choice = new Choice(this, cc["Name"]);
					choice.Probability.val = cc["Probability"].AsFloat;
					choice.Trigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myChoices.Add(choice);
				}

				myChoices.Sort(Choice.SortByNameAscending);
				List<string> options = myChoiceChooser.choices;
				options.Clear();
				for (int e=0; e<myChoices.Count; ++e)
					options.Add(myChoices[e].Name);
				myChoiceChooser.choices = null; // force UI sync
				myChoiceChooser.choices = options;

				if (options.Count == 0)
					myChoiceChooser.valNoCallback = "";
				else if (!options.Contains(myChoiceChooser.val))
					myChoiceChooser.valNoCallback = options[0];
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.Update();
			
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myChoiceChooser.setCallbackFunction(myChoiceChooser.val);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Deregister();
			myChoices.Clear();

			Utils.OnDestroyUI();
		}

		private void TriggerAction()
		{
			if (!myAcceptTrigger.val || myChoices.Count == 0)
				return;

			bool avoidRepetition = myAvoidRepetition.val;
			if (myChoices.Count < 3 || myPreviousIndex1 >= myChoices.Count)
				myPreviousIndex1 = -1;
			if (myChoices.Count < 4 || myPreviousIndex2 >= myChoices.Count)
				myPreviousIndex2 = -1;

			float sum = 0.0f;
			for (int i=0; i<myChoices.Count; ++i)
			{
				if (avoidRepetition && (i == myPreviousIndex1 || i == myPreviousIndex2))
					continue;

				sum += Mathf.Max(0, myChoices[i].Probability.val);
			}
			if (sum == 0.0f)
				return;

			float threshold = UnityEngine.Random.Range(0.0f, sum);
			sum = 0.0f;
			int idx = -1;
			for (int i=0; i<myChoices.Count; ++i)
			{
				if (avoidRepetition && (i == myPreviousIndex1 || i == myPreviousIndex2))
					continue;
				if (myChoices[i].Probability.val > 0)
				{
					idx = i;
					sum += myChoices[i].Probability.val;
					if (sum >= threshold)
						break;
				}
			}

			myPreviousIndex2 = myPreviousIndex1;
			myPreviousIndex1 = idx;
			if (idx >= 0)
				myChoices[idx].Trigger.Trigger();
		}
    }
}
