using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MacGruber;

namespace HAL9001
{
	public class UI_Sync : MVRScript
	{
		// enums would be soo much better... but they crash the VaM compiler
		protected const int TYPE_INVALID = 0;
		protected const int TYPE_BOOL    = 1;
		protected const int TYPE_FLOAT   = 2;
		protected const int TYPE_STRING  = 3;


		protected JSONStorableStringChooser m_atomsJSON;
		protected JSONStorableStringChooser m_storablesJSON;
		protected JSONStorableStringChooser m_parametersJSON;
		
		protected int                m_syncType;
		protected Atom               m_atom;
		protected JSONStorable       m_storable;
		protected JSONStorableParam  m_parameter;

		protected JSONStorableBool   m_boolParam;
		protected Toggle             m_toggle;

		protected JSONStorableFloat  m_floatParam;
		protected Slider             m_slider;

		protected JSONStorableString        m_stringParam;
		protected JSONStorableStringChooser m_stringChooserParam;
		protected Text                      m_text;


		public override void Init()
		{
			try
			{
				if      (containingAtom.type.Equals("UIToggle")) { m_syncType = TYPE_BOOL;   }
				else if (containingAtom.type.Equals("UISlider")) { m_syncType = TYPE_FLOAT;  }
				else if (containingAtom.type.Equals("UIText"  )) { m_syncType = TYPE_STRING; }
				else
				{
					m_syncType = TYPE_INVALID;
					logError("This script needs to be attached to either: a UIToggle atom (for boolean values), a UISlider atom (for float values), or a UIText atom (for string values)");
					this.enabled = false;
					return;
				}

				SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;

				Utils.SetupInfoText(this,
					"<color=#606060><size=40><b>UI_Sync</b></size>\nLogicBrick that synchronises bool/float/string parameters and corresponding UI elements (bool: Toggle, float: Slider, string: Text) without the need for actions or events\n" +
					"Change in either the parameter or the UI element are immediately synchronised.\n</color>",
					455.0f, true
				);

				m_atomsJSON = new JSONStorableStringChooser("Atom", new List<string>{"None"}, "None", "Sync Atom", GatherStorables);
				RegisterStringChooser(m_atomsJSON);
				var atomPopup = CreateFilterablePopup(m_atomsJSON, false);
				atomPopup.popupPanelHeight = 1100f;

				m_storablesJSON = new JSONStorableStringChooser("Storable", new List<string>{"None"}, "None", "Storable", GatherParameters);
				RegisterStringChooser(m_storablesJSON);
				var storablePopup = CreateFilterablePopup(m_storablesJSON, false);
				storablePopup.popupPanelHeight = 1000f;

				m_parametersJSON = new JSONStorableStringChooser("Parameter", new List<string>{"None"}, "None", "Parameter", SynchroniseParameter);
				RegisterStringChooser(m_parametersJSON);
				var parameterPopup = CreateFilterablePopup(m_parametersJSON, false);
				atomPopup.popupPanelHeight = 900f;

				GatherAtoms();
				GatherControl();
			}
			catch (Exception e)
			{
				logError(e.ToString());
			}
		}


		protected void GatherAtoms()
		{
			List<string> atomUIDs = new List<string>();
			foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
			{
				Atom atom = SuperController.singleton.GetAtomByUid(atomUID);
				if (atom != containingAtom)
				{
					atomUIDs.Add(atomUID);
				}
			}
			m_atomsJSON.choices = atomUIDs;
		}


		protected void GatherControl()
		{
			if (m_syncType == TYPE_BOOL)
			{
				foreach (Canvas c in containingAtom.canvases)
				{
					m_toggle = c.GetComponentInChildren<Toggle>(true);
					if (m_toggle != null) break;
				}
				if (m_toggle == null) logError("Could not find Toggle component");
			}
			else if (m_syncType == TYPE_FLOAT)
			{
				foreach (Canvas c in containingAtom.canvases)
				{
					m_slider = c.GetComponentInChildren<Slider>(true);
					if (m_slider != null) break;
				}
				if (m_slider == null) logError("Could not find Slider component");
			}
			else if (m_syncType == TYPE_STRING)
			{
				foreach (Canvas c in containingAtom.canvases)
				{
					m_text = c.GetComponentInChildren<Text>(true);
					if (m_text != null) break;
				}
				if (m_text == null) logError("Could not find Text component");
			}
		}


		protected void GatherStorables(string atomUID)
		{
			List<string> storableIDs = new List<string>();
			if (atomUID != null)
			{
				m_atom = SuperController.singleton.GetAtomByUid(atomUID);
				if (m_atom != null)
				{
					List<string> ids = m_atom.GetStorableIDs();
					foreach (string id in ids)
					{
						JSONStorable s = m_atom.GetStorableByID(id);
						if (s != null) 
						{
							if (    ( (m_syncType == TYPE_FLOAT ) && (s.GetFloatParamNames().Count > 0) )
								 || ( (m_syncType == TYPE_BOOL  ) && (s.GetBoolParamNames().Count  > 0) )
								 || ( (m_syncType == TYPE_STRING) && (s.GetStringParamNames().Count + s.GetStringChooserParamNames().Count > 0) ) )
							{
								storableIDs.Add(id);
							}
						}
					}
				}
			}
			else
			{
				m_atom = null;
			}
			m_storablesJSON.choices = storableIDs;
			m_storablesJSON.val = "None";
		}
		

		protected void GatherParameters(string storableID)
		{
			List<string> parameterNames = new List<string>();
			if ((m_atom != null) && (storableID != null))
			{
				m_storable = m_atom.GetStorableByID(storableID);
				if (m_storable != null)
				{
					if (m_syncType == TYPE_FLOAT ) 
					{ 
						parameterNames.AddRange(m_storable.GetFloatParamNames()); 
					}
					else if (m_syncType == TYPE_BOOL)
					{ 
						parameterNames.AddRange(m_storable.GetBoolParamNames());  
					}
					else if (m_syncType == TYPE_STRING) 
					{ 
						parameterNames.AddRange(m_storable.GetStringParamNames());
						parameterNames.AddRange(m_storable.GetStringChooserParamNames());
					}
				}
			}
			else
			{
				m_storable = null;
			}
			m_parametersJSON.choices = parameterNames;
			m_parametersJSON.val = "None";
		}
		
		
		protected void SynchroniseParameter(string parameterName)
		{
			if ((parameterName != null) && (m_storable != null))
			{
				if (m_syncType == TYPE_BOOL)
				{
					m_boolParam = m_storable.GetBoolJSONParam(parameterName);
					if (m_boolParam != null)
					{
						if (m_toggle != null)
						{
							m_toggle.isOn = m_boolParam.val;
							m_toggle.onValueChanged.AddListener(OnBoolToggleChanged);
						}
						m_boolParam.setCallbackFunction += OnBoolParamChanged;
					}
				}
				else if (m_syncType == TYPE_FLOAT)
				{
					m_floatParam = m_storable.GetFloatJSONParam(parameterName);
					if (m_floatParam != null)
					{
						if (m_slider != null)
						{
							m_slider.minValue = m_floatParam.min;
							m_slider.maxValue = m_floatParam.max;
							m_slider.value    = m_floatParam.val;
							m_slider.onValueChanged.AddListener(OnFloatSliderChanged);
						}
						m_floatParam.setCallbackFunction += OnFloatParamChanged;
					}
				} 
				else if (m_syncType == TYPE_STRING)
				{
					m_stringParam = m_storable.GetStringJSONParam(parameterName);
					if (m_stringParam != null)
					{
						if (m_text != null)
						{
							m_text.text = m_stringParam.val;
						}
						m_stringParam.setCallbackFunction += OnStringParamChanged;
					}
					else
					{
						m_stringChooserParam = m_storable.GetStringChooserJSONParam(parameterName);
						if (m_stringChooserParam != null)
						{
							if (m_text != null)
							{
								m_text.text = m_stringChooserParam.val;
							}
							m_stringChooserParam.setCallbackFunction += OnStringParamChanged;
						}
					}
				}  
			}
		}
		

		protected void OnBoolParamChanged(bool _newValue)
		{
			if (m_toggle != null)
			{
				m_toggle.isOn = _newValue;
			}
		}


		protected void OnBoolToggleChanged(bool _newValue)
		{
			if (m_boolParam != null)
			{
				m_boolParam.val = _newValue;
			}
		}
		

		protected void OnFloatParamChanged(float _newValue)
		{
			if (m_slider != null)
			{
				m_slider.value = _newValue;
			}
		}


		protected void OnFloatSliderChanged(float _newValue)
		{
			if (m_floatParam != null)
			{
				m_floatParam.val = _newValue;
			}
		}
		

		protected void OnStringParamChanged(string _newValue)
		{
			if (m_text != null)
			{
				m_text.text = _newValue;
			}
		}       

		
		protected void OnAtomRename(string _oldUID, string _newUID)
		{
			if ((m_parameter != null) && (m_atom.name.Equals(_oldUID)))
			{
				string storableID    = m_storable.name;
				string parameterName = m_parameter.name;
				GatherAtoms();
				m_atomsJSON.val      = _newUID;
				m_storablesJSON.val  = storableID;
				m_parametersJSON.val = parameterName;
			}
		}


		// ********************************************************************
		#region Logging
		
		protected void logDebug(string text) 
		{
			SuperController.LogMessage($"{nameof(UI_Sync)}: {text}");
		}
		
		protected void logError(string text) 
		{
			SuperController.LogError($"{nameof(UI_Sync)}: {text}");
		}

		#endregion

	
		
		// ********************************************************************
		#region Enable, Disable, Destroy

		public void OnEnable()
		{
			SynchroniseParameter(m_parameter.name);
		}


		public void OnDisable()
		{
			if (m_boolParam != null)
			{
				m_boolParam.setCallbackFunction -= OnBoolParamChanged;
				m_boolParam = null;
			}
			if (m_toggle != null)
			{
				m_toggle.onValueChanged.RemoveListener(OnBoolToggleChanged);
				m_toggle = null;
			}

			if (m_floatParam != null)
			{
				m_floatParam.setCallbackFunction -= OnFloatParamChanged;
				m_floatParam = null;
			}
			if (m_slider != null)
			{
				m_slider.onValueChanged.RemoveListener(OnFloatSliderChanged);
				m_slider = null;
			}

			if (m_stringParam != null)
			{
				m_stringParam.setCallbackFunction -= OnStringParamChanged;
				m_stringParam = null;
			}
			if (m_stringChooserParam != null)
			{
				m_stringChooserParam.setCallbackFunction -= OnStringParamChanged;
				m_stringChooserParam = null;
			}
			if (m_text != null)
			{
				// nothing to do here
				m_text = null;
			}
		}


		void OnDestroy()
		{
			OnDisable();
		}

		#endregion
	}
}