﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using MacGruber;

namespace HAL9001
{
    public class ValueDelay : MVRScript
    {
		private JSONStorableFloat myValue;
		private JSONStorableFloat myIncreaseDelay;
		private JSONStorableFloat myDecreaseDelay;
		private JSONStorableFloat myEffectiveOutput;
		private FloatTrigger      myFloatTrigger;
		private float             currentValue, targetValue, velocity;

        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ValueDelay</b></size>\nLogicBrick that delays float values from a source.\n\n</color>",
				455.0f, true
			);

			myFloatTrigger = new FloatTrigger(this, "OnValueSet");
			Utils.SetupButton(this, "Actions OnValueSet", myFloatTrigger.OpenPanel, false);

			JSONStorableAction resetValue = new JSONStorableAction("Reset", () => { OnResetTrigger();	});
			RegisterAction(resetValue);
	
			myValue = Utils.SetupSliderFloat(this, "Value", 0, 0, 1, false);
			myValue.setCallbackFunction += ValueTrigger;
			currentValue = 0;
			targetValue  = 0;
			velocity     = 0;
			
			myIncreaseDelay = Utils.SetupSliderFloatWithRange(this, "Increase Delay", 0, 0, 600, false);
			myDecreaseDelay = Utils.SetupSliderFloatWithRange(this, "Decrease Delay", 0, 0, 600, false);

			myEffectiveOutput = new JSONStorableFloat("Effective Output", 0.0f, 0.0f, 1.0f);
			UIDynamicSlider slider = CreateSlider(myEffectiveOutput, false);
			slider.quickButtonsEnabled  = false;
			slider.defaultButtonEnabled = false;
			slider.rangeAdjustEnabled   = false;
			slider.sliderControl.slider.interactable = false;
			
			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myFloatTrigger.SyncAtomNames();		
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myFloatTrigger.Name] = myFloatTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myFloatTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myFloatTrigger.Remove();
		}

		private void Update()
		{
			float oldValue = currentValue;
			
			if (currentValue < targetValue)
			{
				currentValue = Mathf.SmoothDamp(currentValue, targetValue, ref velocity, myIncreaseDelay.val, Mathf.Infinity, Time.deltaTime);
			}
			else if (currentValue > targetValue)
			{
				currentValue = Mathf.SmoothDamp(currentValue, targetValue, ref velocity, myDecreaseDelay.val, Mathf.Infinity, Time.deltaTime);
			}
			
			float delta = currentValue - oldValue;
			
			if (delta != 0.0f)
			{
				// close enough?
				if (Mathf.Abs(targetValue - currentValue) > 0.001f)
				{
					// no > update
					myFloatTrigger.Trigger(currentValue);
				}
				else
				{
					// yes > stop all "change"
					OnResetTrigger();
				}
				myEffectiveOutput.val = currentValue;
			}
			
			myFloatTrigger.Update();
		}

		private void ValueTrigger(float v)
		{
			targetValue = v;
		}
		
		private void OnResetTrigger()
		{
			currentValue = targetValue;
			velocity     = 0;
			myFloatTrigger.Trigger(currentValue);
		}
    }
}
